/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2002 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

/* ====================================================================
 * Copyright (c) 2002, The Regents of the University of California.
 *
 * The Regents of the University of California MAKE NO REPRESENTATIONS OR
 * WARRANTIES ABOUT THE SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT. The Regents of the University of California SHALL
 * NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF
 * USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 * ====================================================================
 */
#include <httpd.h>
#include <http_config.h>
#include <http_protocol.h>
#include <http_log.h>
#include <http_request.h>
#include <http_core.h>		/* for ap_construct_url */
#include <mod_dav.h>

#include <apr.h>
#include <apr_strings.h>
#include <apr_hash.h>
#include <apr_tables.h>
#include <apr_file_io.h>

#include "dav_repos.h"
#include "dbms.h"
#include "util.h"
#include "version.h"


/* ### should move these report names to a public header to share with
   ### the client (and third parties). */
static const dav_report_elem avail_reports[] = {
  { "DAV:", "version-tree" },
 
  { NULL },
};

static void dav_repos_get_vsn_options(apr_pool_t *p, apr_text_header *phdr)
{
    /* Note: we append pieces with care for Web Folders's 63-char limit
       on the DAV: header */
    TRACE();
    
    apr_text_append(p, phdr,
		    "version-control,checkout, checkin, report");
    apr_text_append(p, phdr,
		    "uncheckout,version-controlled-collection");
    
}

static dav_error *dav_repos_get_option(const dav_resource *resource,
				       const apr_xml_elem *elem,
				       apr_text_header *option)
{
    TRACE();
    return NULL;
}

/*
** All resource are versionable 
*/
static int dav_repos_versionable(const dav_resource *resource)
{
    TRACE();
    return 1;
}

static dav_auto_version dav_repos_auto_versionable(const dav_resource *resource)
{
     TRACE();
     /*
     return DAV_AUTO_VERSION_ALWAYS;
     */
     /*## FIXME :
       1. checkout
       2. proppatch (try to checkout/proppatch/checkin)
       3. checkin <- fail
     */
     return DAV_AUTO_VERSION_NEVER;
}

/* Put a resource under version control. */
static dav_error *dav_repos_vsn_control(dav_resource *resource,
                                      const char *target)
{
    dav_error *err;
    dav_repos_db *db = resource->info->db;
    dav_repos_resource *db_r = (dav_repos_resource *) resource->info->db_r;
   
    TRACE();
    
    if(target)
	DBG1("TARGET: %s", target);
    
    if ((err = dbms_set_is_versioned(db, db_r, 1)) != NULL)
	return err;

    /* Let's create new VCR */
    err = dbms_create_vr(db, db_r, 0);
    if (err)
	return err;
    
    /* Set initial version */
    db_r->checkin = 1;

    /* Put a resource version control */
    return NULL;
}

static dav_error *dav_repos_checkout(dav_resource *resource,
				     int auto_checkout,
				     int is_unreserved, int is_fork_ok,
				     int create_activity,
				     apr_array_header_t *activities,
				     dav_resource **working_resource)
{
    dav_error *err;
    dav_repos_db *db = resource->info->db;
    dav_repos_resource *db_r = (dav_repos_resource *) resource->info->db_r;
    
    TRACE();
  
    /* Check pre-condition */
    if (db_r->checkin == -1)
	return dav_new_error(db_r->p, HTTP_CONFLICT, 0, 
			     "<DAV:must-be-checked-in/>");
    
    /* Auto-Versioning Stuff */
    if (auto_checkout) {
	return NULL;
    }

    /* Let's set checkin and checkout value */
    err = dbms_set_checkin_out(db, db_r, -1, db_r->checkin);
    if (err)
	return err;
   
    
    /* What should I feed? */
    *working_resource = apr_pcalloc(db_r->p, sizeof(*resource));
    memcpy(*working_resource, resource, sizeof(*resource));
        
    /* mod_dav needs only URI */
    (*working_resource)->uri = 
      dav_repos_mk_vcr_uri(db_r, db_r->checkin, 0);
    
    /* Check post-condition */
    if ((*working_resource)->uri == NULL) 
	return dav_new_error(db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0, 
			     "Can't make VCR URI");
        
    db_r->checkout = db_r->checkin;
    db_r->checkin = -1;

    return NULL;
}

static dav_error *dav_repos_uncheckout(dav_resource *resource)
{
    dav_error *err;
    dav_repos_db *db = resource->info->db;
    dav_repos_resource *db_r = (dav_repos_resource *) resource->info->db_r;
    
    TRACE();
  
    /* Check pre-condition */
    if (db_r->checkout == -1)
	return dav_new_error(db_r->p, HTTP_CONFLICT, 0, 
			     "<DAV:must-be-checked-out/>");
     
    /* Let's set checkin and checkout value */
    err = dbms_set_checkin_out(db, db_r, db_r->checkout, -1);
    if (err)
	return err;
    
    db_r->checkin = db_r->checkout;
    db_r->checkout = -1;

    return NULL;
}


static dav_error *dav_repos_checkin(dav_resource *resource,
                                  int keep_checked_out,
                                  dav_resource **version_resource)
{
    dav_error *err;
    dav_repos_db *db = resource->info->db;
    dav_repos_resource *db_r = (dav_repos_resource *) resource->info->db_r;
    
    TRACE();
  
    /* Check pre-condition */
    if (db_r->checkout == -1)
	return dav_new_error(db_r->p, HTTP_CONFLICT, 0, 
			     "<DAV:must-be-checked-out/>");
    
    /* Let's create new VCR */
    err = dbms_create_vr(db, db_r, db_r->checkout);
    if (err)
	return err;

    /* Let's set checkin and checkout value */
    err = dbms_set_checkin_out(db, db_r, db_r->checkout+1, -1);
    if (err)
	return err;
    
    /* What should I feed? */
    *version_resource = apr_pcalloc(db_r->p, sizeof(*resource));
    memcpy(*version_resource, resource, sizeof(*resource));
    
    /* mod_dav needs only URI */
    (*version_resource)->uri = 
      dav_repos_mk_vcr_uri(db_r, db_r->checkout, 0);

    /* Check post-condition */
    if ((*version_resource)->uri == NULL) 
	return dav_new_error(db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0, 
			     "Can't make VCR URI");
    
    /* FIXME: Do we need this? */
    /* Update number */
    db_r->checkin = db_r->checkout+1;
    db_r->checkout = -1;
    
    return NULL;
} 

static dav_error *dav_repos_avail_reports(const dav_resource *resource,
                                        const dav_report_elem **reports)
{
    TRACE();
 
    /* ### further restrict to the public space? */
    if (resource->type != DAV_RESOURCE_TYPE_REGULAR) {
	*reports = NULL;
	return NULL;
    }

    *reports = avail_reports;
    return NULL;
}

static int dav_repos_report_label_header_allowed(const apr_xml_doc *doc)
{
    TRACE();
    return 0;
}

static void send_xml(apr_bucket_brigade *bb, ap_filter_t *output, 
		     const char *str)
{
    (void) apr_brigade_puts(bb, ap_filter_flush, output, str);
}

static dav_repos_report_elem *dav_repos_build_report_elem(apr_pool_t *pool, 
							  const apr_xml_doc *doc) 
{
    apr_xml_elem *elem;
    dav_repos_report_elem *root=NULL;
    dav_repos_report_elem *ret=NULL, *next;
    TRACE();

    /* Should we check the root is <version-control> ?? */

    elem = dav_find_child(doc->root, "prop");
    if (elem == NULL) {
	DBG0(" NO prop element");
	return root;
    }

    for (elem = elem->first_child; elem; elem = elem->next) {
	dav_repos_report_elem *new = apr_pcalloc(pool, sizeof(*new));
	/* Should I use apr_pstrdup?? */
	new->name = elem->name;
	
	DBG1("Add dav_repos_report_element : %s", new->name);

	/* Add link */
	new->next = root;
	root = new;
    }
    
    /* reverse order */
    while (root) {
	next = root->next;
	root->next = ret;
	ret = root;
	root = next;
    }
    return ret;
}

static dav_error *dav_repos_deliver_report(request_rec *r,
					   const dav_resource *resource,
					   const apr_xml_doc *doc,
					   ap_filter_t *output)
{
    /* this buffers the output for a bit and is automatically flushed,
       at appropriate times, by the Apache filter system. */
    apr_bucket_brigade *bb;
    
    dav_error *err;
    dav_repos_resource *vrs = NULL;
    apr_pool_t *pool = resource->pool;
    dav_repos_report_elem *tmp_elem;
    dav_repos_report_elem *report_elem;
    dav_repos_db *db = resource->info->db;
    dav_repos_resource *db_r = (dav_repos_resource *) resource->info->db_r;
    
    TRACE();
    
    bb =  apr_brigade_create(pool,  /* not the subpool! */
			     output->c->bucket_alloc);
    
    /* Let's find VRS */
    if ((err=dbms_get_vrs(db, db_r, &vrs)) != NULL)
	return err;
    
#if 0 
    /* For test */
    if (vrs==NULL) {
	vnum = db_r->checkout!=-1?
	  db_r->checkout:db_r->checkin;
	
	for (i = 0; i <= vnum; i++) {
	    dav_repos_resource *tmp = 
	      apr_pcalloc(pool, sizeof(*vrs));
	    
	    tmp->root_version = db_r->root_version;
	    tmp->uri = db_r->uri;
	   
	    tmp->version = i;
	    tmp->next = vrs;
	    vrs = tmp;
	}
    }
#endif
    /* 
    ** Read XML and build dav_repos_report_elem
    */
    report_elem = dav_repos_build_report_elem(pool, doc);
    
    if (report_elem == NULL)
	return dav_new_error(pool,
			     HTTP_INTERNAL_SERVER_ERROR, 0, 
			     "No element for report");
    
    send_xml(bb, output, DAV_XML_HEADER);
    send_xml(bb, output, DEBUG_CR);
    send_xml(bb, output, "<D:multistatus xmlns:D=\"DAV:\">" DEBUG_CR);

    /*
    ** vrs should have a pool and uri
    */
    for ( ; vrs; vrs = vrs->next) {
	char *vr_uri;
	int need_404 = 0;

       	/* Set root version */
	vrs->root_version = db_r->uri;
	/* Must have a pool */
	vrs->p = pool;
	/* set last version */
	vrs->lastversion = db_r->checkout!=-1?
	  db_r->checkout-1:db_r->checkin-1;
	
	/* If null, we should skip */
	if ((vr_uri=dav_repos_mk_vcr_uri(db_r, vrs->version, 0))==NULL)
	    continue;
	
	/* Href display */
	send_xml(bb, output, "<D:response>" DEBUG_CR);
	send_xml(bb, output, "<D:href>");
	send_xml(bb, output, vr_uri);
	send_xml(bb, output,  "</D:href>" DEBUG_CR);
	
	/* 
	**   build vpr hash 
	*/
	vrs->type = DAV_RESOURCE_TYPE_VERSION;
	dav_repos_build_vpr_hash(vrs);
	
	/*# TODO: Find props using hash */
	/* Found */
	send_xml(bb, output, "<D:propstat><D:prop>" DEBUG_CR);

	/* We support only DAV namespace, so don't care namespace */
	for (tmp_elem=report_elem; tmp_elem; tmp_elem=tmp_elem->next) {
	    const char *val;
	    
	    /* It shouldn't be a NULL */
	    if (tmp_elem->name==NULL)
		continue;
	    
	    /* Find value from version hash */
	    val = apr_hash_get(vrs->vpr_hash, 
			       tmp_elem->name, APR_HASH_KEY_STRING);
	    if (val==NULL) {
		need_404 = 1;
		tmp_elem->found = 0;
	    } else {
		const char *str = apr_psprintf(pool, "<D:%s>%s</D:%s>" 
					       DEBUG_CR,
					       tmp_elem->name, val, 
					       tmp_elem->name);
		send_xml(bb, output, str);
		tmp_elem->found = 1;
	    }
	}
	send_xml(bb, output, "</D:prop>" DEBUG_CR);
	send_xml(bb, output, "<D:status>HTTP/1.1 200 OK</D:status>" DEBUG_CR);
	send_xml(bb, output, "</D:propstat>" DEBUG_CR);

	/* Not found */
	if (need_404) {
	    send_xml(bb, output, "<D:propstat><D:prop>" DEBUG_CR);
	    
	    /* We support only DAV namespace, so don't care namespace */
	    for (tmp_elem=report_elem; tmp_elem; tmp_elem=tmp_elem->next) {
		if (tmp_elem->name && tmp_elem->found==0) {
		    const char *str = apr_psprintf(pool, "<D:%s/>" DEBUG_CR, 
						   tmp_elem->name);
		    send_xml(bb, output, str);
		}
	    }
	    send_xml(bb, output, "</D:prop>" DEBUG_CR);
	    send_xml(bb, output, "<D:status>HTTP/1.1 404 Not Found</D:status>" 
		     DEBUG_CR);
	    send_xml(bb, output, "</D:propstat>" DEBUG_CR);
	}
	
	send_xml(bb, output, "</D:response>" DEBUG_CR);
  
	/* flush the contents of the brigade */
	ap_fflush(output, bb);
    }
    
    send_xml(bb, output, "</D:multistatus>" DEBUG_CR);
  
    /* flush the contents of the brigade */
    ap_fflush(output, bb);
  
    return NULL;
}


const dav_hooks_vsn dav_repos_hooks_vsn = {
  dav_repos_get_vsn_options,
  dav_repos_get_option,
  dav_repos_versionable,
  dav_repos_auto_versionable,
  dav_repos_vsn_control,
  dav_repos_checkout,
  dav_repos_uncheckout,
  dav_repos_checkin,
  dav_repos_avail_reports,
  dav_repos_report_label_header_allowed,
  dav_repos_deliver_report,
  NULL,                 /* update */
  NULL,                 /* add_label */
  NULL,                 /* remove_label */
  NULL,                 /* can_be_workspace */
  NULL,                 /* make_workspace */
  NULL,                 /* can_be_activity */
  NULL,                 /* make_activity */
  NULL,                 /* merge */
};
